/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/AddT.hpp
 *  \brief  Definition of the templated class GP::AddT.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.3 $
 *  $Date: 2004/04/25 23:12:20 $
 */

/*!
 *  \defgroup MathPrimit Mathematical GP Primitives
 *  \ingroup Primit
 *  \brief Mathematical GP primitives library, part of the GP primitives module.
 */
 
#ifndef Beagle_GP_AddT_hpp
#define Beagle_GP_AddT_hpp

#include <string>

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/AllocatorT.hpp"
#include "beagle/PointerT.hpp"
#include "beagle/ContainerT.hpp"
#include "beagle/WrapperT.hpp"
#include "beagle/GP/Primitive.hpp"
#include "beagle/GP/PrimitiveInline.hpp"
#include "beagle/GP/Context.hpp"
#include "beagle/GP/Datum.hpp"

#ifdef BEAGLE_HAVE_RTTI
#include <typeinfo>
#endif // BEAGLE_HAVE_RTTI


namespace Beagle {
namespace GP {

/*!
 *  \class AddT beagle/GP/AddT.hpp "beagle/GP/AddT.hpp"
 *  \brief GP generic 2 arguments addition primitive class.
 *  \param T Type of the data added. operator+ must be properly defined for this type.
 *  \ingroup GPF
 *  \ingroup Primit
 *  \ingroup MathPrimit
 */
template <class T>
class AddT : public Primitive {

public:

  //! GP::AddT allocator type.
  typedef AllocatorT<AddT<T>,Primitive::Alloc>
          Alloc;
  //! GP::AddT handle type.
  typedef PointerT<AddT<T>,Primitive::Handle>
          Handle;
  //! GP::AddT bag type.
  typedef ContainerT<AddT<T>,Primitive::Bag>
          Bag;

  explicit AddT(std::string inName="ADD");
  virtual ~AddT() { }

  virtual std::string getArgType(unsigned int inN) const;
  virtual std::string getReturnType() const;
  virtual void        execute(GP::Datum& outResult, GP::Context& ioContext);

};


/*!
 *  \brief Add primitive for the double type.
 */
typedef AddT<Double>
        Add;

}
}


/*!
 *  \brief Construct a new addition primitive.
 *  \param inName Name of the addition primitive.
 */
template <class T>
Beagle::GP::AddT<T>::AddT(std::string inName) :
  Beagle::GP::Primitive(2, inName)
{ }


/*!
 *  \brief Return the tag of the type of data needed as input for the primitive AddT.
 *  \param inN Index of the argument to get the type tag.
 *  \return String tagging the data type needed.
 *  \throw AssertException If the index inN given is > to 1.
 *  \throw InternalException If RTTI is disabled and the method is not properly overdefined.
 */
template <class T>
std::string Beagle::GP::AddT<T>::getArgType(unsigned int inN) const
{
#ifdef BEAGLE_HAVE_RTTI
  Beagle_AssertM(inN < 2);
  return typeid(T).name();
#else  // BEAGLE_HAVE_RTTI
  throw Beagle_UndefinedMethodInternalExceptionM("getArgType","GP::AddT");
  return "";
#endif // BEAGLE_HAVE_RTTI
}


/*!
 *  \brief Return the string tagging the type of data returned but the primitive AddT.
 *  \return String tagging the type of data returned.
 *  \throw InternalException If RTTI is disabled and the method is not properly overdefined.
 */
template <class T>
std::string Beagle::GP::AddT<T>::getReturnType() const
{
#ifdef BEAGLE_HAVE_RTTI
  return typeid(T).name();
#else  // BEAGLE_HAVE_RTTI
  throw Beagle_UndefinedMethodInternalExceptionM("getReturnType","GP::AddT");
  return "";
#endif // BEAGLE_HAVE_RTTI
}


/*!
 *  \brief Execute the characteristic operation of the addition primitive.
 *  \param outResult Result of the addition.
 *  \param ioContext Evolutionary context.
 *  \warning operator+ must be properly defined for the type T, the type of data added.
 */
template <class T>
void Beagle::GP::AddT<T>::execute(Beagle::GP::Datum& outResult, Beagle::GP::Context& ioContext)
{
  T& lResult = castObjectT<T&>(outResult);
  get1stArgument(lResult, ioContext);
  T lArg2;
  get2ndArgument(lArg2, ioContext);
  lResult = lResult + lArg2;
}


#endif // Beagle_GP_AddT_hpp
